#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_charset.h"
#include "udm_spell.h"
#include "udm_proto.h"
#include "udm_cache.h"
#include "udm_crc32.h"
#include "udm_indexer.h"

#define LOGDIR	"raw"
#define TREEDIR	"tree"

/******** Convert category string into 32 bit number *************/
static void UdmDecodeCatStr(char * cat_str, int * cat, int * mask){
unsigned int t[5];
char str[128]="";

	strcpy(str,cat_str);
	strcat(str,"000000000000");
	str[10]=0;
	sscanf(str,"%02x%02x%02x%02x%02x",t+0,t+1,t+2,t+3,t+4);

	*mask= (int)( (t[0]?0x7F<<25:0) | (t[1]?0x7F<<18:0) | 
		(t[2]?0x3F<<12:0) | (t[3]?0x3F<<6:0) | (t[4]?0x3F<<0:0));

	*cat= (int)((t[0]<<25) | (t[1]<<18) | (t[2]<<12) | (t[3]<<6) | (t[4]<<0));
}

/*************************** Sort functions **************************/
/* Sort SEARCHWORD by url_id order */
static int cmpurlid(const void *s1,const void *s2){
	return(((UDM_SEARCHWORD*)s1)->url_id-((UDM_SEARCHWORD*)s2)->url_id);
}

/* Function to sort CRCWORD list in URL_ID order */
static int cmpcrc(const void *s1,const void *s2){
int n1,n2;
	n1=(unsigned char)((unsigned int)(((UDM_CRCWORD*)s1)->wrd_id)>>28);
	n2=(unsigned char)((unsigned int)(((UDM_CRCWORD*)s2)->wrd_id)>>28);
	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return(0);
}

/* Function to sort CRCCACHE list in WRD_ID order */
static int cmplog(const void *s1,const void *s2){
unsigned int n1,n2;
	n1=((UDM_LOGWORD*)s1)->wrd_id;
	n2=((UDM_LOGWORD*)s2)->wrd_id;
	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return(0);
}

/* Function to sort CACHEWORD list in URL_ID order */
static int cmpcache(const void *s1,const void *s2){
int n1,n2;
	n1=((UDM_LOGWORD*)s1)->wrd_id;
	n2=((UDM_LOGWORD*)s2)->wrd_id;
	if(n1==n2){
		n1=((UDM_LOGWORD*)s1)->url_id;
		n2=((UDM_LOGWORD*)s2)->url_id;
	}
	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return(0);
}

/* Sort tables by their length order */
static int cmptable(const void * t1, const void * t2){
	return(((UDM_CACHETABLE*)t2)->len-((UDM_CACHETABLE*)t1)->len);
}

/******* These two functions open/close all 256+1 logs ***********/

int UdmOpenCache(int * crcdict, char * errmsg){
int i;
char fname[UDMSTRSIZ];
	for(i=0;i<257;i++){
		if(i<256){
			sprintf(fname,"%s%c%s%c%02X.log",UDM_VAR_DIR,UDMSLASH,LOGDIR,UDMSLASH,i);
		}else{
			sprintf(fname,"%s%c%s%cdel.log",UDM_VAR_DIR,UDMSLASH,LOGDIR,UDMSLASH);
		}
		if((crcdict[i]=open(fname,O_RDWR|O_CREAT|O_APPEND,0644))<0){
			sprintf(errmsg,"Can't open \"%s\" for writing: %s",fname,strerror(errno));
			UdmCloseCache(crcdict);
			return(IND_ERROR);
		}
	}
	return(IND_OK);
}

void UdmCloseCache(int * crcdict){
int i;
	for(i=0;i<257;i++)
		if(crcdict[i]>0)
			close(crcdict[i]);
}

/**************** This function is currently not used ***************/

int UdmDeleteFromCache(char *fname, UDM_LOGWORD *todel, int del_count){
	int oldfd,newfd;
	UDM_CACHEHEADER	header;
	UDM_CACHEHEADER	newheader;
	UDM_CACHETABLE	table[4096];
	UDM_CACHETABLE	newtable[4096];
	UDM_CACHEWORD	*cache;
	int i,data_len;
	char tmpname[UDMSTRSIZ]="";

	if((oldfd=open(fname,O_RDONLY))>=0){
		read(oldfd,&header,sizeof(header));
		read(oldfd,table,header.ntables*sizeof(UDM_CACHETABLE));
		data_len=0;
		for(i=0;i<header.ntables;i++)data_len+=table[i].len;
		cache=(UDM_CACHEWORD*)malloc(data_len);
		read(oldfd,cache,data_len);
		close(oldfd);

		for(i=0;i<header.ntables;i++){
			int pos,len,j,k;
			int last,from;

			pos=table[i].pos/sizeof(UDM_CACHEWORD);
			len=table[i].len/sizeof(UDM_CACHEWORD);

			/* Make holes */
			for(j=pos;j<pos+len;j++){
				for(k=0;k<del_count;k++){
					if(cache[j].url_id==todel[k].url_id){
						cache[j].url_id=0;
						table[i].len-=sizeof(UDM_CACHEWORD);
						break;
					}
				}
			}

			/* Turn off holes */
			if(table[i].len>0){
				j=pos;
				while((cache[j].url_id!=0)&&(j<pos+len))j++;
				last=j;
				while(j<pos+len){
					while((cache[j].url_id==0)&&(j<pos+len))j++;
					from=j;
					while((cache[j].url_id!=0)&&(j<pos+len))j++;
					memcpy(&cache[last],&cache[from],(j-from)*sizeof(UDM_CACHEWORD));
					last+=j-from;
				}
			}
		}

		/* Create new table */
		data_len=0;
		newheader.ntables=0;
		for(i=0;i<header.ntables;i++){
			if(table[i].len>0){
				newtable[newheader.ntables].wrd_id=table[i].wrd_id;
				newtable[newheader.ntables].pos=data_len;
				newtable[newheader.ntables].len=table[i].len;
				data_len+=newtable[newheader.ntables].len;
				newheader.ntables++;
			}
		}

		/* Write new cache file */
		sprintf(tmpname,"%s.tmp",fname);
		newfd=open(tmpname,O_WRONLY|O_CREAT|O_TRUNC,0644);
		write(newfd,&newheader,sizeof(newheader));
		write(newfd,newtable,newheader.ntables*sizeof(UDM_CACHETABLE));
		write(newfd,cache,data_len);
		close(newfd);

		UDM_FREE(cache);
		rename(tmpname,fname);

		return(0);
	}else{
		return(1);
	}
}


/***** Write a marker that the old content of url_id should be deleted ****/
int UdmDeleteURLFromCache(UDM_AGENT * Indexer,int url_id, int * crcdict){
UDM_LOGWORD cw;
	cw.url_id=url_id;
	cw.wrd_id=0;
	cw.category=0;
	if((sizeof(cw)!=write(crcdict[256],&cw,sizeof(cw)))){
		sprintf(UdmAgentErrorMsg(Indexer),"Can't write to delete log: %s",strerror(errno));
		UdmCloseCache(crcdict);
		return(IND_ERROR);
	}
	return(IND_OK);
}


/******** Write words into 256+1 logs **************************/
int UdmStoreWordsCache(UDM_AGENT * Indexer,int url_id,char * catstr,int * crcdict){
int fd,i,bytes;
int category,mask;
UDM_CRCWORD * cw;

	/* Checke whether XX.log files are already open */
	if(!crcdict[0]){
		/* Not open yet, let's open it now */
		i=UdmOpenCache(crcdict,UdmAgentErrorMsg(Indexer));
		if(i!=IND_OK){
			UdmCloseCache(crcdict);
			return(i);
		}
	}
	/* Convert category string into 32 bit number */
	UdmDecodeCatStr(catstr,&category,&mask);

	/* Mark that old content should be deleted */
	i=UdmDeleteURLFromCache(Indexer,url_id,crcdict);
	if(i!=IND_OK)return(i);

	if(Indexer->nwords){
		int prev=-1,first=0;
		bytes=sizeof(UDM_CRCWORD)*Indexer->nwords;
		cw=(UDM_CRCWORD*)malloc(bytes);
		for(i=0;i<Indexer->nwords;i++){
			cw[i].url_id=url_id;
			cw[i].weight=Indexer->Word[i].count,
			cw[i].wrd_id=UdmStrCRC32(Indexer->Word[i].word);
			//printf("%08X\t%d\t%s\n",cw[i].wrd_id,url_id,Indexer->Word[i].word);
		}
		qsort((void*)cw,Indexer->nwords,sizeof(UDM_CRCWORD),cmpcrc);
		for(i=0;i<Indexer->nwords;i++){
			int n;
			n=((unsigned int)cw[i].wrd_id)>>24;
			if((n!=prev)||(i+1==Indexer->nwords)){
				if(first!=i){
					int w,wcount;
					UDM_LOGWORD * buf;
					fd=crcdict[prev];
					wcount=i-first;
					bytes=wcount*sizeof(UDM_LOGWORD);
					buf=(UDM_LOGWORD *)malloc(bytes);
					for(w=0;w<wcount;w++){
						//printf("%08X %08X %02X %02X\n",cw[first+w].url_id,cw[first+w].wrd_id,prev,cw[first+w].weight);
						buf[w].category=category;
						buf[w].url_id=cw[first+w].url_id;
						buf[w].wrd_id=(cw[first+w].wrd_id<<8)+(cw[first+w].weight&0xFF);
						//printf("%08X %08X\n",buf[w].url_id,buf[w].wrd_id);
					}
					if((bytes!=write(fd,(void *)buf,bytes))){
						free(buf);
						sprintf(UdmAgentErrorMsg(Indexer),"Can't write to log file #%d: %s",prev,strerror(errno));
						UdmCloseCache(crcdict);
						return(IND_ERROR);
					}
					free(buf);
				}
				first=i;
			}
			prev=n;
		}
		free(cw);
	}	
	return(IND_OK);
}

/***************** Slit one log into cache *****************/

int UdmSplitCacheLog(int log){
int fd,bytes,i,bufsize;
char fname[UDMSTRSIZ]="";
UDM_LOGWORD * buf;

	/* Open log file */
	sprintf(fname,"%s%c%s%c%02X.log",UDM_VAR_DIR,UDMSLASH,LOGDIR,UDMSLASH,log);
	if((fd=open(fname,O_RDONLY))<0){
		return(IND_ERROR);
	}
	/* Allocate buffer */
	bufsize=sizeof(UDM_LOGWORD)*1024*1024*16;
	buf=(UDM_LOGWORD*)malloc(bufsize);

	while((bytes=read(fd,buf,bufsize))){
		int n,new,prev,first=0;
		n=bytes/sizeof(UDM_LOGWORD);
		printf("Load %d/%d bytes\n",bytes,bufsize);
		qsort(buf,n,sizeof(UDM_LOGWORD),cmplog);
		prev=((unsigned int)buf[0].wrd_id)>>8;

		for(i=1;i<n+1;i++){
			if(i<n)new=((unsigned int)buf[i].wrd_id)>>8;
			else	new=prev+1;

			if((i==n)||((((unsigned int)prev)>>12)!=(((unsigned int)new)>>12))){
				if(first!=i){
					int newfd,oldfd;
					char scrc[16];
					char dname[UDMSTRSIZ]="";
					char tmpname[UDMSTRSIZ]="";
					int r,w,t,t_prev,t_new,pos,count=0;
					UDM_CACHEHEADER	header;
					UDM_CACHETABLE	table[4096];
					UDM_CACHEWORD cache[4096];
					UDM_LOGWORD    * logwords = NULL;
						
					sprintf(scrc,"%08X",(((unsigned int)prev)|(((unsigned int)log)<<24))&0xFFFFF000);
					sprintf(dname,"%s%c%s%c%02X%c%c%c",UDM_VAR_DIR,UDMSLASH,TREEDIR,UDMSLASH,log,UDMSLASH,scrc[2],UDMSLASH);
					UdmBuild(dname,0755);
					sprintf(dname,"%s%c%s%c%02X%c%c%c",UDM_VAR_DIR,UDMSLASH,TREEDIR,UDMSLASH,log,UDMSLASH,scrc[2],UDMSLASH);
						
					//sprintf(fname,"%s%c%c",dname,scrc[3],scrc[4]);
					sprintf(fname,"%s%s",dname,scrc);
					printf("%s new: %08X prev: %08X first: %d i: %d\n",fname,new,prev,first,i);

					if((oldfd=open(fname,O_RDONLY))>=0){
						int j;
							
						printf("Read old: %s\n",fname);
						read(oldfd,&header,sizeof(header));
						read(oldfd,table,header.ntables*sizeof(UDM_CACHETABLE));
							
						for(w=0;w<header.ntables;w++){
							int c=0;
							int n=table[w].len/sizeof(UDM_CACHEWORD);
							while((r=(n-c))>0){
								if(r>4096)r=4096;
								read(oldfd,cache,r*sizeof(UDM_CACHEWORD));
								if(!count)logwords=(UDM_LOGWORD*)malloc((count+r)*sizeof(UDM_LOGWORD));
								else	logwords=(UDM_LOGWORD*)realloc(logwords,(count+r)*sizeof(UDM_LOGWORD));
								for(j=0;j<r;j++){
									logwords[count+j].wrd_id=table[w].wrd_id;
									logwords[count+j].url_id=cache[j].url_id;
									logwords[count+j].category=cache[j].category;
								}
								c+=r;
								count+=r;
							}
							//printf("w=%d count=%d\n",w, count);
						}
						close(oldfd);
					}
						
					r=i-first;
					if(!count)logwords=(UDM_LOGWORD*)malloc((count+r)*sizeof(UDM_LOGWORD));
					else	logwords=(UDM_LOGWORD*)realloc(logwords,(count+r)*sizeof(UDM_LOGWORD));
					memcpy(logwords+count,buf+first,r*sizeof(UDM_LOGWORD));
					count+=r;
					
					qsort(logwords,count,sizeof(UDM_LOGWORD),cmpcache);

					/* Create table */
					header.version=0;
					header.ntables=0;
					pos=0;
					t_prev=logwords[0].wrd_id;
					for(t=1;t<count+1;t++){
						if(t<count)t_new=logwords[t].wrd_id;
						else	t_new=t_prev+1;
						if(t_prev!=t_new){
							table[header.ntables].wrd_id=t_prev;
							table[header.ntables].pos=pos;
							table[header.ntables].len=t*sizeof(UDM_CACHEWORD)-pos;
							pos+=table[header.ntables].len;
							header.ntables++;
						}
						t_prev=t_new;
					}
						
					sprintf(tmpname,"%s%c%c.tmp",dname,scrc[3],scrc[4]);
					newfd=open(tmpname,O_WRONLY|O_CREAT|O_TRUNC,0644);
					write(newfd,&header,sizeof(header));
					write(newfd,table,header.ntables*sizeof(UDM_CACHETABLE));
					
					w=0;
					while(w<count){
						int j;
						if((count-w)>4096)r=4096;
						else	r=(count-w);
						
						for(j=0;j<r;j++){
							cache[j].url_id=logwords[w+j].url_id;
							cache[j].category=logwords[w+j].category;
						}
						w+=r;
						write(newfd,cache,r*sizeof(UDM_CACHEWORD));
					}
					close(newfd);
					UDM_FREE(logwords);
					rename(tmpname,fname);
				}
				first=i;
			}
			prev=new;
		}
	}
	close(fd);
	UDM_FREE(buf);
	return IND_OK;
}



/****************** Search stuff ************************************/

UDM_SEARCHWORD * UdmFindCache(UDM_AGENT * Indexer,char *q){
char * w, * lasttok;
char * words[100];
int    cwords[100];
int i,j,mask,cat;
int num_words=0;
UDM_SEARCHWORD * wrd = NULL;
UDM_SEARCHWORD * wrd1 = NULL;

typedef struct m {
	int last;
	int cur;
} UDM_MERG;

UDM_MERG merg[1024];

typedef struct pm {
	UDM_SEARCHWORD * plast;
	UDM_SEARCHWORD * pcur;
} UDM_PMERG;

UDM_PMERG pmerg[1024];

register int nmerg=0;
register int min_num=0;
register int nwrd=0;
register int nwrd1=0;

#ifdef DEBUG_SEARCH
unsigned long ticks;
#endif

	num_words=0;
	Indexer->wordinfo[0]='\0';
	w=UdmGetWord(q,&lasttok,Indexer->charset);
	while(w){
		char ** ww;
		char *rw;

		ww=UdmNormalizeWord(Indexer,w);
		rw=ww?*ww:w;

		/*
		if(UdmIsStopWord(rw)){
			sprintf(UDM_STREND(wordinfo)," %s :stopword",rw);
			w=UdmGetWord(NULL,&lasttok,LCharset);
			continue;
		}
		*/

		if(num_words<100){
			words[num_words]=strdup(rw);
			cwords[num_words]=UdmStrCRC32(rw);
			(num_words)++;
		}
		w=UdmGetWord(NULL,&lasttok,Indexer->charset);
	}
	
	UdmDecodeCatStr(Indexer->Conf->category,&cat,&mask);

#ifdef DEBUG_SEARCH
	fprintf(stderr,"Start read cache %d words\n",num_words);
	ticks=UdmStartTimer();
#endif
	for(i=0;i<num_words;i++){
		int fd1,t;
		char scrc[16];
		char dname[UDMSTRSIZ];
		char fname[UDMSTRSIZ];
		
		UDM_CACHEHEADER  header;
		UDM_CACHETABLE   table[8192];
		UDM_CACHEWORD  * cw = NULL;
		int data_ofs;
		int group_num=0;
		int group_first=0;
		int group;

		sprintf(scrc,"%08X",cwords[i]&0xFFFFF000);
		sprintf(dname,"%s%c%s%c%c%c%c%c%c",
			UDM_VAR_DIR,UDMSLASH,TREEDIR,UDMSLASH,
			scrc[0],scrc[1],UDMSLASH,scrc[2],UDMSLASH);
		sprintf(fname,"%s%s",dname,scrc);
		if((fd1=open(fname,O_RDONLY))<0){
#ifdef DEBUG_SEARCH
			fprintf(stderr,"\tCan't open: %s\n",fname);
#endif
			continue;
		}
#ifdef DEBUG_SEARCH
	fprintf(stderr,"\tOpened: %s\n",fname);
#endif
		read(fd1,&header,sizeof(header));
		data_ofs=sizeof(header)+sizeof(UDM_CACHETABLE)*header.ntables;
		read(fd1,table,header.ntables*sizeof(UDM_CACHETABLE));
		qsort(&table,header.ntables,sizeof(UDM_CACHETABLE),cmptable);
		
		group_first=nwrd;
		group=0;
		for(t=0;t<header.ntables;t++){
			int bytes;
			if((cwords[i]&0x00FFFFFF)==(((unsigned int)(table[t].wrd_id))>>8)){
				int num,real_num=0;
				
				cw=(UDM_CACHEWORD*)malloc(table[t].len);
				lseek(fd1,data_ofs+table[t].pos,SEEK_SET);
				bytes=read(fd1,cw,table[t].len);
				if(bytes==-1)fprintf(stderr,"Error reading cache file %s\n",fname);
				num=bytes/sizeof(UDM_CACHEWORD);   
				if(!nwrd)wrd=(UDM_SEARCHWORD*)malloc(num*sizeof(UDM_SEARCHWORD));
				else	wrd=(UDM_SEARCHWORD*)realloc(wrd,(num+nwrd)*sizeof(UDM_SEARCHWORD));
				
				for(j=0;j<num;j++){
					if((cw[j].category&mask)==cat){
						wrd[nwrd+real_num].url_id=cw[j].url_id;
						wrd[nwrd+real_num].count=1;
						wrd[nwrd+real_num].weight=table[t].wrd_id&0xFF;
						real_num++;
					}
				}
				if(group<2){
					merg[nmerg].cur=nwrd;
					merg[nmerg].last=nwrd+real_num;
					nmerg++;
				}
				group_num+=real_num;
				nwrd+=real_num;
				group++;
			}
		}
		close(fd1);

		if((group==0)&&(Indexer->search_mode==UDM_MODE_ALL)){
			Indexer->total_found=0;
			return(NULL);
		}
	
		sprintf(UDM_STREND(Indexer->wordinfo)," %s: %d",words[i],group_num);

		merg[nmerg-1].last=nwrd;

		qsort(wrd+merg[nmerg-1].cur,
			merg[nmerg-1].last-merg[nmerg-1].cur,
			sizeof(UDM_SEARCHWORD),cmpurlid);

	}
#ifdef DEBUG_SEARCH
	ticks=UdmStartTimer() - ticks;
	fprintf(stderr,"Stop read cache:\t%.2f\n",(float)ticks/1000);
	fprintf(stderr,"Start merge %d groups\n",nmerg);
	ticks=UdmStartTimer();
#endif
	wrd1=(UDM_SEARCHWORD*)malloc(nwrd*sizeof(UDM_SEARCHWORD));

	for(i=0;i<nmerg;i++){
		pmerg[i].pcur=wrd+merg[i].cur;
		pmerg[i].plast=wrd+merg[i].last;
	}

	while(nmerg){
		register int k, min_val=0x7FFFFFFF;
		register UDM_PMERG * p;

		if(nmerg==1){
			int num;
			
			num=pmerg[0].plast-pmerg[0].pcur;
			memcpy(&wrd1[nwrd1],pmerg[0].pcur,num*sizeof(UDM_SEARCHWORD));
			nwrd1+=num;
			break;//nmerg=0;
		}else{
			for(k=0;k<nmerg;k++){
				if(min_val>(pmerg[k].pcur)->url_id){
					min_val=(pmerg[k].pcur)->url_id;
					min_num=k;	
				}
			}
			p=pmerg+min_num;
			memcpy(&wrd1[nwrd1],p->pcur,sizeof(UDM_SEARCHWORD));
			p->pcur++;
			if(p->pcur==p->plast){
				nmerg--;
				if(min_num<nmerg){
					memmove(p,p+1,sizeof(UDM_PMERG)*(nmerg-min_num));
				}
			}
			nwrd1++;
		}
	}
	UDM_FREE(wrd);
	
	Indexer->total_found=nwrd1;
#ifdef DEBUG_SEARCH
	ticks=UdmStartTimer() - ticks;
	fprintf(stderr,"Stop merge groups:\t%.2f nwrd=%d nwrd1=%d\n",(float)ticks/1000,nwrd,nwrd1);
#endif
	
	return(wrd1);
}
